package consume

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"log"
	"query-service/internal/entity"
	"query-service/pkg/errorhandler"
	"time"

	"git.science.uu.nl/datastrophe/broker"
	"git.science.uu.nl/datastrophe/query-conversion"
)

// HandleMessage gets called when a message is received
func (s *Service) HandleMessage(msg *broker.Message) {
	// Grab the headers from the message
	sessionID, clientID, queryID, err := getHeaders(&msg.Headers)
	if err != nil {
		log.Println("no headers in message")
		return
	}

	s.publishStatus("Received", sessionID, queryID)

	// Unmarshall the incoming message into an IncomingJSONQuery object
	JSONQuery, err := query.UnmarshalJSON(&msg.Body)
	if err != nil {
		s.publishError(sessionID, queryID, entity.MalformedRequestError, "malformed request")
		return
	}

	// Check if there was a database name
	if JSONQuery.DatabaseName == "" {
		// If not, there is no way to proceed past this point
		s.publishError(sessionID, queryID, entity.MalformedRequestError, "malformed request - no database name")
		return
	}

	s.publishStatus("Translating", sessionID, queryID)

	// Convert the json byte msg to a query string
	query, err := s.queryConverter.ConvertQuery(JSONQuery)
	if err != nil {
		s.publishError(sessionID, queryID, entity.TranslationError, "failed to translate query")
		return
	}

	// Send the resulting query back to the client
	translationMsg := entity.MessageStruct{
		QueryID: *queryID,
		Type:    "query_translation_result",
		Value:   *query,
	}
	translationMsgBytes, _ := json.Marshal(translationMsg)

	s.producer.PublishMessage(&translationMsgBytes, sessionID)

	s.publishStatus("Getting database credentials", sessionID, queryID)

	// Fetch database credentials from the user service
	databaseInfo, err := s.databaseInfoService.GetDatabaseInfo(clientID, &JSONQuery.DatabaseName)
	if err != nil {
		s.publishError(sessionID, queryID, entity.DatabaseCredentialsError, "failed to retrieve database credentials")
		return
	}

	s.publishStatus("Processing on database", sessionID, queryID)

	// execute and retrieve result
	// convert result to general (node-link) format
	result, err := s.requestSender.ExecuteQuery(*query, databaseInfo.Username, databaseInfo.Password, databaseInfo.URL, databaseInfo.Port, databaseInfo.InternalDatabaseName)
	if err != nil {
		s.publishError(sessionID, queryID, entity.DatabaseExecutionError, "query failed to execute on database")
		return
	}

	s.publishStatus("Completed", sessionID, queryID)

	// Add type indicator to result from database
	var res interface{}
	json.Unmarshal(*result, &res)
	resultMsg := entity.MessageStruct{
		QueryID: *queryID,
		Type:    "query_result",
		Value:   res,
	}
	resultMsgBytes, err := json.Marshal(resultMsg)
	if err != nil {
		errorhandler.LogError(err, "Marshalling query_result went wrong!") // TODO: send error message to client instead
		return
	}

	// Publish the result
	s.producer.PublishMessage(&resultMsgBytes, sessionID)

	// Caching message in object store
	if *queryID != "" {
		cacheQueryResultContext, cancelCacheQueryResult := context.WithTimeout(context.Background(), time.Second*20)
		defer cancelCacheQueryResult()
		s.objectStore.Put(cacheQueryResultContext, "cached-queries", fmt.Sprintf("%s-%s", *sessionID, *queryID), int64(len(resultMsgBytes)), bytes.NewReader(resultMsgBytes))
	}
}
